/* 
 * DANG_BEGIN_MODULE
 *
 * Description: Port handling code for DOSEMU
 * 
 * Maintainers: Volunteers?
 *
 * This is the code that allows and disallows port access within DOSEMU.
 * The BOCHS port IO code was actually very cleverly done.  So I stole
 * the idea.
 *
 * DANG_END_MODULE
 *
 * $Date: $
 * $Source: $
 * $Revision: $
 * $State: $
 *
 * $Log: $
 */

#include <unistd.h>
#include <malloc.h>

#include "dosemu.h"
#include "emu_bios.h"
#include "emu_hgc.h"
#include "emu_video.h"
#include "emu_serial.h"
#include "emu_port.h"
#include "emu_iodev.h"
#include "emu_termio.h"

/* PORT_DEBUG is to specify whether to record port writes to debug output.
 * 0 means disabled.
 * 1 means record all port accesses to 0x00 to 0xFF
 * 2 means record ANY port accesses!  (big fat debugfile!)
 */ 
#define PORT_DEBUG 2


#define MAX_IO_DEVICES  32

static struct {
  Bit8u   (* read_portb)(Bit32u port_addr);
  void    (* write_portb)(Bit32u port_addr, Bit8u byte);
  Bit16u  (* read_portw)(Bit32u port_addr);
  void    (* write_portw)(Bit32u port_addr, Bit16u word);
  char * handler_name;
  int    irq;
} port_handler[EMU_MAX_IO_DEVICES];

static Bit8u port_handler_id[0x10000];     /* 64K ports */
static Bit8u port_handles;                 /* number of io_handler's */
static char *irq_handler_name[EMU_MAX_IRQS];

#define emu_handler(port)    port_handler[port_handler_id[port]]

/* 
 * DANG_BEGIN_FUNCTION port_inb(Bit32u port)
 *
 * description:
 *   Handles/simulates an inb() port IO read
 *
 * DANG_END_FUNCTION
 */
Bit8u port_inb(Bit32u port)
{
  Bit8u res;
  port &= 0xffff;
  res   = emu_handler(port).read_portb(port);

#if PORT_DEBUG > 0
#if PORT_DEBUG == 1
  if (port < 0x100)
#endif
    i_printf("PORT: \"%s\" inb [0x%04lx] -> 0x%02hx\n", 
	     emu_handler(port).handler_name, port, res);
#endif
  return res;
}

/* 
 * DANG_BEGIN_FUNCTION port_outb(Bit32u port, Bit8u byte)
 *
 * description:
 *   Handles/simulates an outb() port IO write
 *
 * DANG_END_FUNCTION
 */
void port_outb(Bit32u port, Bit8u byte)
{
  port &= 0xffff;
  emu_handler(port).write_portb(port, byte);

#if PORT_DEBUG > 0
#if PORT_DEBUG == 1
  if (port < 0x100)
#endif
    i_printf("PORT: \"%s\" outb [0x%04lx] <- 0x%02hx\n",
	     emu_handler(port).handler_name, port, byte);
#endif
}

/* 
 * DANG_BEGIN_FUNCTION port_inw(Bit32u port)
 *
 * description:
 *   Handles/simulates an inw() port IO read.  Usually this invokes
 *   port_inb() twice, but it may be necessary to do full word i/o for
 *   some video boards.
 *
 * DANG_END_FUNCTION
 */
Bit16u port_inw(Bit32u port)
{
  Bit16u res;
  port &= 0xffff;

  if (emu_handler(port).read_portw != NULL) {
    res = emu_handler(port).read_portw(port);
#if PORT_DEBUG > 0
#if PORT_DEBUG == 1
    if (port < 0x100)
#endif
      i_printf("PORT: \"%s\" inw [0x%04lx] -> 0x%04hx\n",
	       emu_handler(port).handler_name, port, res);
#endif
  }
  else {
    res = (Bit16u)port_inb(port) | (((Bit16u)port_inb(port+1)) << 8);
  }

  return res;
}

/* 
 * DANG_BEGIN_FUNCTION port_outw(Bit32u port, Bit16u word)
 *
 * description:
 *   Handles/simulates an outw() port IO write
 *
 * DANG_END_FUNCTION
 */
void port_outw(Bit32u port, Bit16u word)
{
  port &= 0xffff;

  if (emu_handler(port).write_portw != NULL) {
    emu_handler(port).write_portw(port, word);
#if PORT_DEBUG > 0
#if PORT_DEBUG == 1
    if (port < 0x100)
#endif
      i_printf("PORT: \"%s\" outw [0x%04lx] <- 0x%04x\n",
	       emu_handler(port).handler_name, port, word);
#endif
  }
  else {
    port_outb(port, word & 0xff);
    port_outb(port+1, (word >> 8) & 0xff);
  }
}

static Bit8u port_default_inb(Bit32u port)
{
  return 0;
}

static void port_default_outb(Bit32u port, Bit8u byte)
{
}

Bit8u port_safe_inb(Bit32u port)
{
  Bit8u res = 0;

  i_printf("PORT: safe_inb ");
  if (i_am_root) {
    iopl(3);
    res = port_in(port);
    iopl(0);
  }
  else
    i_printf("want to ");
  i_printf("in(%lx)", port);
  if (i_am_root)
    i_printf(" = %hx", res);
  i_printf("\n");
  return res;
}

void port_safe_outb(Bit32u port, Bit8u byte)
{
  i_printf("PORT: safe_outb ");
  if (i_am_root) {
    iopl(3);
    port_out(byte, port);
    iopl(0);
  }
  else
    i_printf("want to ");
  i_printf("outb(%lx, %hx)\n", port, byte);
}

Bit16u port_safe_inw(Bit32u port)
{
  Bit16u res = 0;

  i_printf("PORT: safe_inw ");
  if (i_am_root) {
    iopl(3);
    res = port_in_w(port);
    iopl(0);
  }
  else
    i_printf("want to ");
  i_printf("inw(%lx)", port);
  if (i_am_root)
    i_printf(" = %x", res);
  i_printf("\n");
  return res;
}

void port_safe_outw(Bit32u port, Bit16u word)
{
  i_printf("PORT: safe_outw ");
  if (i_am_root) {
    iopl(3);
    port_out_w(word, port);
    iopl(0);
  }
  else
    i_printf("want to ");
  i_printf("outw(%lx, %x)\n", port, word);
}

/* 
 * DANG_BEGIN_FUNCTION emu_port_init(void)
 *
 * description:
 *   Resets all the port port_handler information
 *
 * DANG_END_FUNCTION
 */
void port_init(void)
{
  int i;

  /* handle 0 maps to the unmapped IO device handler.  Basically any
     ports which don't map to any other device get mapped to this
     handler which does absolutely nothing.
   */
  port_handler[0].read_portb     = port_default_inb;
  port_handler[0].write_portb    = port_default_outb;
  port_handler[0].read_portw     = NULL;
  port_handler[0].write_portw    = NULL;
  port_handler[0].handler_name   = "unknown port";
  port_handler[0].irq            = EMU_NO_IRQ;

  port_handler[1].read_portb     = port_safe_inb;
  port_handler[1].write_portb    = port_safe_outb;
  port_handler[1].read_portw     = NULL;
  port_handler[1].write_portw    = NULL;
  port_handler[1].handler_name   = "safe port io";
  port_handler[1].irq            = EMU_NO_IRQ;

  port_handles = 2;

  /* set unused elements to appropriate values */
  for (i=1; i < MAX_IO_DEVICES; i++) {
    port_handler[i].read_portb   = NULL;
    port_handler[i].write_portb  = NULL;
    port_handler[i].read_portw   = NULL;
    port_handler[i].write_portw  = NULL;
  }

  for (i=0; i < 0x10000; i++)
    port_handler_id[i] = 0;  /* unmapped IO handle */
}

void port_register_handler(emu_iodev_t device)
{
  int handle, i;

  i_printf("PORT: registering \"%s\" for port range [0x%04x-0x%04x]\n",
	   device.handler_name, device.start_addr, device.end_addr);

  if (device.irq != EMU_NO_IRQ  &&  device.irq >= EMU_MAX_IRQS) {
    error("PORT: IO device %s registered with IRQ=%d above %u\n",
	  device.handler_name, device.irq, EMU_MAX_IRQS-1);
    leavedos(1);
  }

  /* first find existing handle for function or create new one */
  for (handle=0; handle < (int) port_handles; handle++) {
    if (!strcmp(port_handler[handle].handler_name, device.handler_name))
      break;
  }

  if (handle >= (int) port_handles) {
    /* no existing handle found, create new one */
    if (port_handles >= EMU_MAX_IO_DEVICES) {
      error("PORT: too many IO devices, increase EMU_MAX_IO_DEVICES\n");
      leavedos(1);
    }

    if (device.irq != EMU_NO_IRQ  &&  irq_handler_name[device.irq]) {
      error("PORT: IRQ %d conflict.  IO devices %s & %s\n",
        device.irq, irq_handler_name[device.irq], device.handler_name);
      leavedos(1);
    }
    if (device.irq != EMU_NO_IRQ  &&  device.irq < EMU_MAX_IRQS)
      irq_handler_name[device.irq] = device.handler_name;
    port_handles++;
    port_handler[handle].read_portb     = device.read_portb;
    port_handler[handle].write_portb    = device.write_portb;
    port_handler[handle].read_portw     = device.read_portw;
    port_handler[handle].write_portw    = device.write_portw;
    port_handler[handle].handler_name   = device.handler_name;
    port_handler[handle].irq            = device.irq;
    }

  /* change table to reflect new handler id for that address */
  for (i=device.start_addr; i <= device.end_addr; i++) {
    if (port_handler_id[i] != 0) {
      error("PORT: IO device address conflict at IO address 0x%x\n", i);
      error("      conflicting devices: %s & %s\n",
	    port_handler[handle].handler_name, emu_handler(i).handler_name);
      leavedos(1);
    }
    port_handler_id[i] = handle;
  }
}












static struct port_struct {
  int start;
  int size;
  int permission;
  int ormask, andmask;
} *ports = NULL;

static int num_ports = 0;

/* control the io permissions for ports */
int
allow_io(int start, int size, int permission, int ormask, int andmask)
{

  /* first the simplest case...however, this simplest case does
   * not apply to ports above 0x3ff--the maximum ioperm()able
   * port under Linux--so we must handle some of that ourselves.
   */
  if (permission == IO_RDWR && (ormask == 0 && andmask == 0xFFFF)) {
    if ((start + size - 1) <= 0x3ff) {
      c_printf("giving hardware access to ports 0x%x -> 0x%x\n",
	       start, start + size - 1);
      set_ioperm(start, size, 1);
      return (1);
    }

    /* allow direct I/O to all the ports that *are* below 0x3ff
       * and add the rest to our list
       */
    else if (start <= 0x3ff) {
      warn("PORT: s-s: %d %d\n", start, size);
      set_ioperm(start, 0x400 - start, 1);
      size = start + size - 0x400;
      start = 0x400;
    }
  }

  /* we'll need to add an entry */
  if (ports)
    ports = (struct port_struct *)
      realloc(ports, sizeof(struct port_struct) * (num_ports + 1));

  else
    ports = (struct port_struct *)
      malloc(sizeof(struct port_struct) * (num_ports + 1));

  num_ports++;
  if (!ports) {
    error("ALLOW_IO: allocation error for ports!\n");
    num_ports = 0;
    return (0);
  }

  ports[num_ports - 1].start = start;
  ports[num_ports - 1].size = size;
  ports[num_ports - 1].permission = permission;
  ports[num_ports - 1].ormask = ormask;
  ports[num_ports - 1].andmask = andmask;

  c_printf("added port %x size=%d permission=%d ormask=%x andmask=%x\n",
	   start, size, permission, ormask, andmask);

  return (1);
}



#if 0



extern int s3_8514_base;

#ifdef PORT_DEBUG
#if PORT_DEBUG == 1
#define DUMP_PORT(str,port,val)  {if (port < 0x100) i_printf(str, port, val);}
#else
#define DUMP_PORT(str,port,val)  i_printf(str, port, val)
#endif
#endif

#define PORT_NULL   0xffffffff

static u_char video_port_io = 0;

/* find a matching entry in the ports array */
static int
find_port(int port, int permission)
{
  static int last_found = 0;
  int i;

  for (i = last_found; i < num_ports; i++)
    if (port >= ports[i].start &&
	port < (ports[i].start + ports[i].size) &&
	((ports[i].permission & permission) == permission)) {
      last_found = i;
      return (i);
    }

  for (i = 0; i < last_found; i++)
    if (port >= ports[i].start &&
	port < (ports[i].start + ports[i].size) &&
	((ports[i].permission & permission) == permission)) {
      last_found = i;
      return (i);
    }

/*
   for now, video_port_io flag allows special access for video regs
*/
  if (   ((LWORD(cs) & 0xf000) == config.vbios_seg) 
      && config.allowvideoportaccess 
      && scr_state.current) {
#if 0
    v_printf("VID: checking port %x\n", port);
#endif
    if (port > 0x4000)
      return (-1);
    port &= 0xfff;
    if (    (port > 0x300 && port < 0x3f8) 
         ||  port > 0x400
         || port == 0x42 || port == 0x1ce || port == 0x1cf) {
      video_port_io = 1;
      return 1;
    }
  }

  return (-1);
}

static inline int
port_readable(int port)
{
  return (find_port(port, IO_READ) != -1);
}

static inline int
port_writeable(int port)
{
  return (find_port(port, IO_WRITE) != -1);
}

int
read_port(int port)
{
  int r;
  int i = find_port(port, IO_READ);

  if (i == -1)
    return (0);

  if (port <= 0x3ff)
    set_ioperm(port, 1, 1);
  else
    iopl(3);

  r = port_in(port);

  if (port <= 0x3ff)
    set_ioperm(port, 1, 0);
  else
    iopl(0);

  if (!video_port_io) {
    r &= ports[i].andmask;
    r |= ports[i].ormask;
  }
  else
    video_port_io = 0;

  h_printf("read port 0x%x gave %02x at %04x:%04x\n",
	   port, r, LWORD(cs), LWORD(eip));
  return (r);
}

int
read_port_w(int port)
{
  int r;
  int i = find_port(port,IO_READ);
  int j = find_port(port+1,IO_READ);

  if(i == -1) {
    i_printf("can't read low-byte of 16 bit port 0x%04x:trying to read high\n",
              port); 
    return(read_port(port+1) << 8 );
  }
  if(j == -1) {
    i_printf("can't read hi-byte of 16 bit port 0x%04x:trying to read low\n",
              port);
  return(read_port(port));
  }

  if(port <= 0x3fe) {
     set_ioperm(port  ,2,1);
  }
  else
     iopl(3);

  r = port_in_w(port);
  if(port <= 0x3fe) {
     set_ioperm(port,2,0);
  }
  else
    iopl(0);

  if(!video_port_io) {
    r &= ( ports[i].andmask | 0xff00 );
    r &= ( (ports[j].andmask << 8) | 0xff);
    r |= ( ports[i].ormask & 0xff ); 
    r |= ( (ports[j].ormask << 8) & 0xff00 );
  }
  else
    video_port_io = 0;

  i_printf("read 16 bit port 0x%x gave %04x at %04x:%04x\n",
           port, r, LWORD(cs), LWORD(eip));
   return (r);
}

int
write_port(int value, int port)
{
  int i = find_port(port, IO_WRITE);

  if (i == -1)
    return (0);

  if (!video_port_io) {
    value &= ports[i].andmask;
    value |= ports[i].ormask;
  }
  else
    video_port_io = 0;

  h_printf("write port 0x%x value %02x at %04x:%04x\n",
	   port, value, LWORD(cs), LWORD(eip));

  if (port <= 0x3ff)
    set_ioperm(port, 1, 1);
  else
    iopl(3);

  port_out(value, port);

  if (port <= 0x3ff)
    set_ioperm(port, 1, 0);
  else
    iopl(0);

  return (1);
}

int
write_port_w(int value,int port)
{
  int i = find_port(port, IO_WRITE);
  int j = find_port(port + 1, IO_WRITE);

  if( (i == -1) || (j == -1) ) {
    i_printf("can't write to 16 bit port 0x%04x\n",port);
    return 0;
  } 
  if(!video_port_io) {
    value &= ( ports[i].andmask | 0xff00 );
    value &= ( (ports[j].andmask << 8) | 0xff );
    value |= ( ports[i].ormask & 0xff );
    value |= ( (ports[j].ormask << 8) & 0xff00);
  }
  else
    video_port_io = 0;

  i_printf("write 16 bit port 0x%x value %04x at %04x:%04x\n",
           port, value, LWORD(cs), LWORD(eip));

  if (port <= 0x3fe) {
    set_ioperm(port  ,2,1);
  }
  else
    iopl(3);

  port_out_w(value, port);

  if (port <= 0x3fe) {
    set_ioperm(port, 2, 0);
  }
  else
    iopl(0);

  return (1);
}

/*
 * DANG_BEGIN_FUNCTION video_port_emuin(int port)
 *
 * description:
 *  Simulates a read from the specified video port
 *
 * DANG_END_FUNCTION
 */
int video_port_emuin(int port)
{
  static unsigned int cga_r = 0;
  unsigned int r = PORT_NULL;

  switch (port) {
  case 0x3da:
    /* graphic status - many programs will use this port to sync with
     * the vert & horz retrace so as not to cause CGA snow */
    i_printf("3ba/3da port inb\n");
    r = (cga_r ^= 1) ? 0xcf : 0xc6;
    break;
    
  case 0x03b4:		/* adr-reg */
  case 0x03b5:		/* data-reg */
  case 0x03ba:		/* status-reg */
    if (config.usesX) {
      v_printf("HGC Portread: %d\n", (int) port);
      r = safe_port_in_byte(port);
    }
    else {
      i_printf("3ba/3da port inb\n");
      r = (cga_r ^= 1) ? 0xcf : 0xc6;
    }
    break;
    
    case 0x03b8:		/* mode-reg */
    if (config.usesX) {
      v_printf("HGC Portread: %d\n", (int) port);
      r = safe_port_in_byte(port);
      r = (r & 0x7f) | (hgc_Mode & 0x80);
    }
    break;
    
  case 0x3bc:
    i_printf("printer port inb [0x3bc] = 0\n");    /* 0 by default */
    break;
    
  case 0x03bf:		/* conf-reg */
    if (config.usesX) {
      v_printf("HGC Portread: %d\n", (int) port);
      r = safe_port_in_byte(port);
      r = (r & 0xfd) | (hgc_Konv & 0x02);
    }
    break;
    
  case 0x3db:			/* light pen strobe reset, 0 by default */
    break;
  }
  return r;
}

/*
 * DANG_BEGIN_FUNCTION video_port_emuout(int port, int byte)
 *
 * description:
 *  Simulates a write to the specified video port
 *
 * DANG_END_FUNCTION
 */
void video_port_emuout(int port, int byte)
{
  switch (port) {
  case 0x03b8:		/* mode-reg */
    if (config.usesX) {
      if (byte & 0x80)
	set_hgc_page(1);
      else
	set_hgc_page(0);
      set_ioperm(port, 1, 1);
      port_out(byte & 0x7f, port);
      set_ioperm(port, 1, 0);
    }
    break;
    
  case 0x03ba:		/* status-reg */
    if (config.usesX) {
      set_ioperm(port, 1, 1);
      port_out(byte, port);
      set_ioperm(port, 1, 0);
    }
    break;

  case 0x03bf:		/* conf-reg */
    if (config.usesX) {
      if (byte & 0x02)
	map_hgc_page(1);
      else
	map_hgc_page(0);
      set_ioperm(port, 1, 1);
      port_out(byte & 0xFD, port);
      set_ioperm(port, 1, 0);
    }
    break;
  }
}

#if 0
/*
 * DANG_BEGIN_FUNCTION port_inb(int port)
 *
 * description:
 *  INB reads a byte from the specified port.
 *
 * DANG_END_FUNCTION
 */
int port_inb(int port)
{
  unsigned int r = PORT_NULL;
  int tmp;

  port &= 0xffff;

#if 0
  if ((port >= 0x00 && port <= 0x1f) || (port >= 0x80 && port <= 0x83) ||
      (port >= 0xc0 && port <= 0xdf)) {
    i_printf("PORT: ERROR DMA not supported (port read: 0x%02x)\n", port);
    return 0;
  }

  if ((port >= 0x90 && port <= 0x97) || (port >= 0x100 && port <= 0x107)) {
    i_printf("PORT: ERROR POS not emulated (port read: 0x%02x)\n", port);
    return 0;
  }

  if ((port >= 0x170 && port <= 0x17f) || (port >= 0x1f0 && port <= 0x1ff)) {
    i_printf("PORT: ERROR can not access disk controller (port read: 0x%02x)\n", port);
    return 0;
  }

  if ((port >= 0x370 && port <= 0x377) || (port >= 0x3f0 && port <= 0x3f7)) {
    i_printf("PORT: ERROR can not access floppy controller (port read: 0x%02x)\n", port);
    return 0;
  }

  if ((port >= 0xf0 && port <= 0xff)) {
    i_printf("PORT: ERROR can not access math coprocessor (port read: 0x%02x)\n", port);
    return 0;
  }
#endif

  if (port >= 0x20 && port <= 0x21) {
    return read_pic0(port-0x20);
  }

  if (port >= 0x40 && port <= 0x47) {
    return pit_read(port);
  }

  if (port >= 0x60 && port <= 0x6f) {
    return keyb_io_read(port);
  }

  if (port >= 0x70 && port <= 0x71) {
    return cmos_read(port);
  }
      
  if (port >= 0xa0 && port <= 0xa1) {
    return read_pic1(port-0xa0);
  }

  if (port >= 0x3b3 && port <= 0x3df) {
    if (config.chipset && config.mapped_bios)
      r = video_port_in(port);             /* vga enabled */
    else
      r = video_port_emuin(port);        /* emulating video */
    if (r != PORT_NULL) {
      DUMP_PORT("PORT: INB video read 0x%04x -> 0x%02x\n", port, r);
      return r;
    }
  }

  /* SERIAL PORT I/O.  The base serial port must be a multiple of 8. */
  for (tmp = 0; tmp < config.num_ser; tmp++)
    if ((port & ~7) == com[tmp].base_port) {
      r = do_serial_in(tmp, port);
      DUMP_PORT("PORT: INB serial read 0x%04x -> 0x%02x\n", port, r);
      return r;
    }

  if ((config.chipset == S3) && ((port & 0x03fe) == s3_8514_base) &&
      (port & 0xfc00)) {
    r = read_port(port) & 0xff;
    DUMP_PORT("PORT: INB video S3 read 0x%04x -> 0x%02x\n", port, r);
    return r;
  }

  /* The diamond bug */
  if (config.chipset == DIAMOND && (port >= 0x23c0) && (port <= 0x23cf)) {
    r = read_port(port) & 0xff;
    DUMP_PORT("PORT: INB video Diamond read 0x%04x -> 0x%02x\n", port, r);
    return r;
  }

  if (port_readable(port)) {
    r = read_port(port) & 0xff;
    DUMP_PORT("PORT: INB read 0x%04x -> 0x%02x\n", port, r);
  }
  else {
    r = 0;
    DUMP_PORT("PORT: INB unavailable 0x%04x -> 0x%02x\n", port, 0);
  }

  return r;
}


/*
 * DANG_BEGIN_FUNCTION port_inw(int port)
 *
 * description:
 *  INW reads a byte from the specified port and another from port+1.
 *
 * DANG_END_FUNCTION
 */
int port_inw(int port)
{
  unsigned int r = PORT_NULL;

  port &= 0xffff;

  if ((config.chipset == S3) && ((port & 0x03ff) == s3_8514_base) &&
      (port & 0xfc00)) {
    r = read_port_w(port) & 0xffff;
    DUMP_PORT("PORT: INW video S3 0x%04x -> 0x%04x\n", port, r);
    return r;
  }

  r = read_port_w(port);
  DUMP_PORT("PORT: INW 0x%04x -> 0x%04x\n", port, r);
  return r;
}

/*
 * DANG_BEGIN_FUNCTION port_outb(int port, int byte)
 *
 * description:
 *  OUTB writes a byte from to specified port.
 *
 * DANG_END_FUNCTION
 */
void port_outb(int port, int byte)
{
  int tmp;

i_printf("PORT: --outb(0x%04x, 0x%02x)\n", port, byte);

  port &= 0xffff;
  byte &= 0xff;

  if ((port >= 0x00 && port <= 0x1f) || (port >= 0x80 && port <= 0x83) ||
      (port >= 0xc0 && port <= 0xdf)) {
    i_printf("PORT: ERROR DMA not supported (port write: 0x%02x, 0x%02x)\n", port, byte);
    return;
  }

  if ((port >= 0x90 && port <= 0x97) || (port >= 0x100 && port <= 0x107)) {
    i_printf("PORT: ERROR POS not emulated (port write: 0x%02x, 0x%02x)\n", port, byte);
    return;
  }

  if ((port >= 0x170 && port <= 0x17f) || (port >= 0x1f0 && port <= 0x1ff)) {
    i_printf("PORT: ERROR can not access disk controller (port write: 0x%02x, 0x%02x)\n", port, byte);
    return;
  }

  if ((port >= 0x370 && port <= 0x377) || (port >= 0x3f0 && port <= 0x3f7)) {
    i_printf("PORT: ERROR can not access floppy controller (port write: 0x%02x, 0x%02x)\n", port, byte);
    return;
  }

  if ((port >= 0xf0 && port <= 0xff)) {
    i_printf("PORT: ERROR can not access math coprocessor (port write: 0x%02x, 0x%02x)\n", port, byte);
    return;
  }

  if (port >= 0x20 && port <= 0x21) {
    write_pic0(port-0x20, byte);
    return;
  }

  if (port >= 0x40 && port <= 0x47) {
    pit_write(port, byte);
    return;
  }

  if (port >= 0x60 && port <= 0x6f) {
    keyb_io_write(port, byte);
    return;
  }

  if (port >= 0x70 && port <= 0x71) {
    cmos_write(port, byte);
    return;
  }
      
  if (port >= 0xa0 && port <= 0xa1) {
    write_pic1(port-0xa0, byte);
    return;
  }

  /*
   * Scott NOTE:  I am unclear on why this code needs to be here
   *              if we're in { console vga } mode...
   */

  /* Port writes for enable/disable blinking character mode */
  if (port == 0x03C0) {
    static int last_byte = -1;
    static int last_index = -1;
    static int flip_flop = 1;

    DUMP_PORT("PORT: Blinking char mode (port write: 0x%02x, 0x%02x)\n", port, byte);
    flip_flop = !flip_flop;
    if (flip_flop) {
      if (last_index = 0x10)
	char_blink = (byte & 8) ? 1 : 0;
      last_byte = byte;
    }
    else {
      last_index = byte;
    }
    return;
  }

  /* Port writes for cursor position */
  if ((port & 0xfffe) == READ_WORD(BIOS_VIDEO_PORT)) {
    /* Writing to the 6845 */
    static int last_port;
    static int last_byte;
    static int hi = 0, lo = 0;
    int pos;

    DUMP_PORT("PORT: Cursor POS (port write: 0x%02x, 0x%02x)\n", port, byte);
    if (!config.usesX) {
      if ((port == READ_WORD(BIOS_VIDEO_PORT) + 1) &&
	  (last_port == READ_WORD(BIOS_VIDEO_PORT))) {
	/* We only take care of cursor positioning for now. */
	/* This code should work most of the time, but can
	     be defeated if i/o permissions change (e.g. by a vt
	     switch) while a new cursor location is being written
	     to the 6845. */
	if (last_byte == 14) {
	  hi = (unsigned char) byte;
	  pos = (hi << 8) | lo;
	  cursor_col = pos % 80;
	  cursor_row = pos / 80;
	  if (config.usesX)
	    poshgacur(cursor_col, cursor_row);
	}
	else if (last_byte == 15) {
	  lo = (unsigned char) byte;
	  pos = (hi << 8) | lo;
	  cursor_col = pos % 80;
	  cursor_row = pos / 80;
	  if (config.usesX)
	    poshgacur(cursor_col, cursor_row);
	}
      }
      last_port = port;
      last_byte = byte;
    }
    else {
      set_ioperm(port, 1, 1);
      port_out(byte, port);
      set_ioperm(port, 1, 0);
    }
    return;
  }

  /*
   * END QUESTIONABLE CODE
   */

  if (port >= 0x3b3 && port <= 0x3df) {
    if (config.chipset && config.mapped_bios)
      video_port_out(port, byte);          /* vga enabled */
    else
      video_port_emuout(port, byte);       /* emulating video */
    DUMP_PORT("PORT: OUTB video write 0x%04x = 0x%02x\n", port, byte);
    return;
  }

  /* SERIAL PORT I/O.  The base serial port must be a multiple of 8. */
  for (tmp = 0; tmp < config.num_ser; tmp++)
    if ((port & ~7) == com[tmp].base_port) {
      do_serial_out(tmp, port, byte);
      DUMP_PORT("PORT: OUTB serial write 0x%04x = 0x%02x\n", port, byte);
      return;
    }

  if ((config.chipset == S3) && ((port & 0x03fe) == s3_8514_base) &&
      (port & 0xfc00)) {
    iopl(3);
    port_out(byte, port);
    iopl(0);
    DUMP_PORT("PORT: OUTB video S3 write 0x%04x = 0x%02x\n", port, byte);
    return;
  }

  /* The diamond bug */
  if ((config.chipset == DIAMOND) && (port >= 0x23c0) && (port <= 0x23cf)) {
    iopl(3);
    port_out(byte, port);
    iopl(0);
    DUMP_PORT("PORT: OUTB video Diamond write 0x%04x = 0x%x\n", port, byte);
    return;
  }

  if (port_writeable(port)) {
    write_port(port, byte);
    DUMP_PORT("PORT: OUTB write 0x%04x = 0x%02x\n", port, byte);
  }
  else {
    DUMP_PORT("PORT: OUTB ignored 0x%04x = 0x%02x\n", port, byte);
  }
}

/*
 * DANG_BEGIN_FUNCTION port_outw(int port, int word)
 *
 * description:
 *  OUTW writes a byte to the specified port and another from port+1.
 *
 * DANG_END_FUNCTION
 */
void port_outw(int port, int value)
{
i_printf("PORT: --outw(0x%04x, 0x%04x)\n", port, value);

  if ((config.chipset == S3) && ((port & 0x03ff) == s3_8514_base) &&
      (port & 0xfc00)) {
    iopl(3);
    port_out_w(value, port);
    iopl(0);
    v_printf("S3 outw [0x%04x] = 0x%04x\n", port, value);
    return;
  }
  if (!write_port_w(value,port)) {
    port_outb(port, value & 0xff);
    port_outb(port + 1, value >> 8);
  }
}

#endif
#endif
